import threading
import socket
import queue
from typing import Callable, Optional, List, Dict, Tuple
import time

from . import logger
from . import broadcaster
from . import msg_parser
from . import receiver
from . import sender

log = logger.Logger("controller")

class SwarmNet:
  def __init__(self, 
               mapping: Dict[str, Callable[[Optional[str]], None]], 
               device_retries: int = 3, 
               device_refresh_interval: int = 60,
               port: int = 51000,
               device_list: List[Tuple[str, int]] = []):
    self.fn_map = mapping
    self.discovery_retries = device_retries
    self.discovery_interval = device_refresh_interval
    self.port = port
    
    if(device_list != []):
      self.fixed_list = True
      self.swarm_list: List[Tuple[str, int]] = device_list
      log.info("Static device list provided")
    else:
      self.fixed_list = False
      self.swarm_list: List[Tuple[str, int]] = []
    
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(("8.8.8.8", 1))
    self.addr = s.getsockname()[0]
    s.close()
    
    log.info(f"This address is {self.addr}:{self.port}")
    
    log.success("SwarmNet controller started")
    
  def start(self) -> None:
    self.swarm_list_lock = threading.Lock()
    self.received_ids = []
    self.received_ids_lock = threading.Lock()
    self.rx_queue = queue.Queue(128)
    self.tx_queue = queue.Queue(32)
    self.fn_map["JOIN"] = self._register_new_member
    self.parser = msg_parser.MessageParser(self.fn_map, self.rx_queue)
    self.receiver = receiver.Receiver(self.addr, self.port, self.add_device, self.has_seen_message, self.append_seen_messages, rx_queue=self.rx_queue, tx_queue=self.tx_queue)
    self.sender = sender.Sender(self.addr, self.tx_queue, self.remove_device)
    self.broadcaster = broadcaster.Broadcaster(self.addr, self.port, self.rx_queue, self.add_device, self.fixed_list, self.swarm_list)
    
    self.parse_thread = threading.Thread(target=parse_thread_target, args=[self])
    self.parse_thread_exit_request = False
    self.parse_thread.start()
    log.info("Parser thread started")
    
    self.receiver_thread = threading.Thread(target=receiver_thread_target, args=[self])
    self.receiver_thread_exit_request = False
    self.receiver_thread.start()
    log.info("Receiver thread started")
    
    self.sender_thread = threading.Thread(target=sender_thread_target, args=[self])
    self.sender_thread_exit_request = False
    self.sender_thread.start()
    log.info("Sender thread started")
    
    log.info_header("Beginning broadcast of JOIN command")
    self.broadcast(f"JOIN {self.addr} {self.port}")
    
  def kill(self) -> None:
    self.broadcaster_thread_exit_request = True
    self.parse_thread_exit_request = True
    self.receiver_thread_exit_request = True
    self.sender_thread_exit_request = True
    
    self.parse_thread.join()
    self.receiver_thread.join()
    self.sender_thread.join()
    
    log.warn("All threads have been killed")
    
  def _register_new_member(self, msg: Optional[str]) -> None:
    addr = msg.split(" ", 1)[0]
    port = msg.split(" ", 1)[1]
    self.add_device((addr, int(port)))
      
  def get_devices(self) -> List[Tuple[str, int]]:
    self.swarm_list_lock.acquire()
    ds = self.swarm_list
    self.swarm_list_lock.release()
    return ds
  
  def set_devices(self, ds: List[Tuple[str, int]]) -> None:
    if(self.fixed_list):
      return
    
    self.swarm_list_lock.acquire()
    self.swarm_list = ds
    self.swarm_list_lock.release()
    
  def add_device(self, d: Tuple[str, int]) -> None:
    if(self.fixed_list):
      return
    
    if(d[0] == self.addr):
      return
    self.swarm_list_lock.acquire()
    
    if(d in self.swarm_list):
      return
    
    self.swarm_list.append(d)
    self.swarm_list_lock.release()
    log.info(f"New device added at {d[0]}:{d[1]}")
    
  def remove_device(self, d: Tuple[str, int]) -> None:
    if(self.fixed_list):
      return
    
    self.swarm_list_lock.acquire()
    self.swarm_list.remove(d)
    self.swarm_list_lock.release()
    log.warn(f"Device removed at {d[0]}:{d[1]}")
    
  def get_seen_messages(self) -> List[str]:
    self.received_ids_lock.acquire()
    ms = self.received_ids
    self.received_ids_lock.release()
    return ms
  
  def set_seen_messages(self, ms: List[str]) -> None:
    self.received_ids_lock.acquire()
    self.received_ids = ms
    self.received_ids_lock.release()
  
  def append_seen_messages(self, m: str) -> None:
    self.received_ids_lock.acquire()
    self.received_ids.append(m)
    self.received_ids_lock.release()
    
  def has_seen_message(self, m: str) -> bool:
    self.received_ids_lock.acquire()
    b = m in self.received_ids
    self.received_ids_lock.release()
    return b
  
  def set_log_level(lv: logger.Logger.Log_Level) -> None:
    log.set_log_level(lv)
    
  def _calc_header(self) -> str:
    return f"{time.time()}/{self.addr}/{self.port}"
    
  def send(self, msg: str):
    header = self._calc_header()
    self.append_seen_messages(header)
    self.tx_queue.put(f"{header}:{msg}", block=True)
    
  def broadcast(self, msg: str):
    header = self._calc_header()
    self.append_seen_messages(header)
    self.broadcaster.broadcast(f"{header}:{msg}")
    
def parse_thread_target(ctrl: SwarmNet):
  while(not ctrl.parse_thread_exit_request):
    if not ctrl.rx_queue.empty():
      ctrl.parser.parse_msg()
    else:
      time.sleep(0.01)
  log.warn("Parse thread killed")
    
def receiver_thread_target(ctrl: SwarmNet):
  while(not ctrl.receiver_thread_exit_request):
    ctrl.receiver.accept_connection()
  log.warn("Receiver thread killed")

    
def sender_thread_target(ctrl: SwarmNet):
  while(not ctrl.sender_thread_exit_request):
    if not ctrl.tx_queue.empty():
      ctrl.sender.flush_queue(ctrl.get_devices())
    else:
      time.sleep(0.01)
  log.warn("Sender thread killed") 
    



import threading
import queue
from typing import Optional, Callable, Dict

from . import logger

log = logger.Logger("parser")

class MessageParser:
  def __init__(self, mapping: Dict[str, Callable[[Optional[str]], None]], q: queue.Queue):
    self.fn_map = mapping
    self.msg_queue = q
    log.success("SwarmNet parser started")
  
  def parse_msg(self) -> None:
    msg: str = self.msg_queue.get(block=True).split(' ', 1)
    cmd: str = msg[0]
    data: Optional[str] = None
    
    if len(msg) > 1:
      data = msg[1]
      
    try:
      # self.ctrl.fn_map[cmd](data)
      threading.Thread(target=self.fn_map[cmd], args=[data]).start()
      log.info(f"Parse thread spawned to parse {cmd} command")
      
    except KeyError:
      log.error(f"No registered parser for command {cmd}")


from enum import Enum
import datetime

class Logger:
  class Colour_Code(Enum):
    ENDC      = '\033[0m'
    BOLD      = '\033[1m'
    UNDERLINE = '\033[4m'
    RED       = '\033[91m'
    GREEN     = '\033[92m'
    YELLOW    = '\033[93m'
    BLUE      = '\033[94m'
    PURPLE    = '\033[95m'
    CYAN      = '\033[96m'
      
  class Log_Level(Enum):
    INFO     = 0
    SUCCESS  = 1
    WARN     = 2
    ERROR    = 3
    CRITICAL = 4
    
  log_level = Log_Level.INFO

  def __init__(self, name):
    self.name = name
      
  def _log_normal(self, s: str, c: Colour_Code) -> None:
    print(f"{c.value}[{self.name}: {datetime.datetime.now()}] {s}{self.Colour_Code.ENDC.value}")
    
  def _log_bold(self, s: str, c: Colour_Code) -> None:
    print(f"{self.Colour_Code.BOLD.value}{c.value}[{self.name}: {datetime.datetime.now()}] {s}{self.Colour_Code.ENDC.value}")
    
  def _log_underlined(self, s: str, c: Colour_Code) -> None:
    print(f"{self.Colour_Code.UNDERLINE.value}{c.value}[{self.name}: {datetime.datetime.now()}] {s}{self.Colour_Code.ENDC.value}")
    
  def _log_bold_underlined(self, s: str, c: Colour_Code) -> None:
    print(f"{self.Colour_Code.BOLD.value}{self.Colour_Code.UNDERLINE.value}{c.value}[{self.name}: {datetime.datetime.now()}] {s}{self.Colour_Code.ENDC.value}")
    
  def info(self, s: str) -> None:
    if(self.log_level.value <= self.Log_Level.INFO.value):
      self._log_normal(s, self.Colour_Code.BLUE)
    
  def info_header(self, s: str) -> None:
    if(self.log_level.value <= self.Log_Level.INFO.value):
      self._log_bold(s, self.Colour_Code.CYAN)
    
  def success(self, s: str) -> None:
    if(self.log_level.value <= self.Log_Level.SUCCESS.value):
      self._log_bold_underlined(s, self.Colour_Code.GREEN)
    
  def warn(self, s: str) -> None:
    if(self.log_level.value <= self.Log_Level.WARN.value):
      self._log_bold(s, self.Colour_Code.YELLOW)
    
  def error(self, s: str) -> None:
    if(self.log_level.value <= self.Log_Level.ERROR.value):
      self._log_bold(s, self.Colour_Code.RED)
    
  def critical(self, s: str) -> None:
    if(self.log_level.value <= self.Log_Level.CRITICAL.value):
      self._log_bold_underlined(s, self.Colour_Code.RED)
      
      
def set_log_level(l: Logger.Log_Level):
  Logger.log_level = l


import socket
import queue
import threading
from typing import Callable, Tuple, List

from . import logger

log = logger.Logger("broadcaster")

TIMEOUT = 0.5

class Broadcaster:
  def __init__(self, self_addr: str, port: int, rx: queue.Queue, add_device_fn: Callable[[Tuple[str, int]], None], fixed_list: bool, devices: List[Tuple[str, int]]):
    self.self_addr = self_addr
    self.port = port
    self.rx = rx
    self.add_device = add_device_fn
    self.fixed_list = fixed_list
    
    if(fixed_list):
      self.device_list = devices
    
    log.success("SwarmNet broadcaster started")
    
  def broadcast(self, msg: str) -> None:
    if(self.fixed_list):
      self._bc_static_list(msg)
    else:
      lower = threading.Thread(target=self._bc_lower, args=[msg])
      upper = threading.Thread(target=self._bc_upper, args=[msg])
      
      lower.start()
      upper.start()
      
      lower.join()
      upper.join()
        
    log.success(f"Broadcast message {msg} sent")
    
  def _bc_static_list(self, msg) -> None:
    for ap in self.device_list:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(TIMEOUT)
        try:
          sock.connect(ap)
          sock.send(bytes(msg + "\n", "utf-8"))
        except OSError:
          pass
        sock.close()
    
  def _bc_lower(self, msg) -> None:
    for addr_end in range(2,128):
        if(str(addr_end) == self.self_addr.split(".")[3]):
          continue
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(TIMEOUT)
        try:
          sock.connect((f"192.168.0.{addr_end}", self.port))
          sock.send(bytes(msg + "\n", "utf-8"))
          self.add_device((f"192.168.0.{addr_end}", self.port))
        except OSError:
          pass
        sock.close()
    
  def _bc_upper(self, msg) -> None:
    for addr_end in range(128,255):
        if(str(addr_end) == self.self_addr.split(".")[3]):
          continue
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(TIMEOUT)
        try:
          sock.connect((f"192.168.0.{addr_end}", self.port))
          sock.send(bytes(msg + "\n", "utf-8"))
          self.add_device((f"192.168.0.{addr_end}", self.port))
        except OSError:
          pass
        sock.close()




import queue
import socket
from typing import Callable, Tuple

from . import logger

log = logger.Logger("receiver")

class Receiver:
  def __init__(self, addr: str, port: int, add_device: Callable[[Tuple[str, int]], None], received: Callable[[str], bool], register_received: Callable[[str], None], rx_queue: queue.Queue, tx_queue: queue.Queue):
    self.listener = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    self.listener.bind((addr, port))
    self.listener.listen()
    self.listener.settimeout(5)
    self.add_device = add_device
    self.received = received
    self.register_received = register_received
    self.rx = rx_queue
    self.tx = tx_queue
  
    log.info("Waiting for connections")
    log.success("SwarmNet receiver started")
    
  def control_receiver(self, exit_request: bool) -> None:
    if(exit_request):
      self.listener.shutdown()
    
  def accept_connection(self) -> None:
    try:
      client_sock, client_info = self.listener.accept()
    except TimeoutError:
      return
    
    log.info(f"Connection received")

    full_data: str = ""
    try:
      while True:
        data = client_sock.recv(1024)
        if not data:
          break
        full_data += data.decode("utf-8")
    except OSError:
      pass
    
    parts = full_data.split(":", 1)
    
    ip = parts[0].split("/", 2)[1]
    port = parts[0].split("/", 2)[2]
    
    self.add_device((ip, int(port)))
    
    if not self.received(parts[0]):
      self.rx.put(parts[1])
      self.register_received(parts[0])
      self.tx.put(full_data) # Gossip protocol


import socket
import queue
from typing import List, Tuple, Callable

from . import logger

log = logger.Logger("sender")

class Sender:
  def __init__(self, self_addr: str, tx_queue: queue.Queue, remove_device_fn: Callable[[Tuple[str, int]], None]):
    self.tx = tx_queue
    self.self_addr = self_addr
    self.remove_dev = remove_device_fn
    log.success("SwarmNet sender started")
    
  def flush_queue(self, devices: List[Tuple[str, int]]):
    while(not self.tx.empty()):
      msg = self.tx.get()
      log.info(f"Sending message: {msg}")
      for (addr, port) in devices:
        if(addr == self.self_addr.split(".")[3]):
          continue
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(10)
        try:
          sock.connect((addr, port))
          sock.send(bytes(msg + "\n", "utf-8"))
        except OSError:
          log.error(f"Failed to send message to {addr}")
          self.remove_dev((addr, port))
        sock.close()
      log.success("Message sent to all swarm devices")
